// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_EXCEPT_H
#define OB_EXCEPT_H

namespace CORBA
{

//
// Undefine minor (this is needed for SunOS 4.1 and some other Unix
// flavors)
//
#undef minor

//
// Declare Any
//
class Any;

//
// The Exception classes
//
class Exception
{
protected:

    Exception() { }

public:

    //
    // Standard IDL to C++ Mapping
    //
    Exception(const Exception&) { }
    virtual ~Exception() { }
    Exception& operator=(const Exception&) { return *this; }

    virtual void _raise() const = 0;

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    virtual Exception* _OB_clone() const = 0;
    virtual void _OB_insertInAny(Any&) = 0;
};

enum CompletionStatus
{
    COMPLETED_YES,
    COMPLETED_NO,
    COMPLETED_MAYBE
};

//
// When adding a new system exception, don't forget to add (or to add
// to) the following stuff:
//
// Except.h:
//   OB_MAKE_STANDARD_EXCEPTION
//
// Except.cpp:
//   OB_MAKE_STANDARD_EXCEPTION_I
//   OB::UnmarshalSystemException, OB::UnmarshalSystemExceptionThrow
//
// Any.h:
//   OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION
//
// Any.cpp:
//   OB_MAKE_STANDARD_EXCEPTION_INSERTION_EXTRACTION_I
//
// Typecode.h:
//   OB_IMPORT extern const OB::TypeCodeSysExConst _tc_*;
//
// Typecode.cpp:
//   const OB::TypeCodeSysExConst _tc_*("*");
//   bool CORBA::TypeCode::_OB_isSystemException() const
//
// Util.cpp:
//   OB::ExceptionToString
//
// Also update the test suite:
//
// Test2.idl:
//   Add void op_*_Ex()
//
// TestDSI_impl.cpp:
//   Add new case for operation
//
// TestStubException.cpp:
//   Add new call for exception
//
// Test_impl.cpp:
//   Add new method
//
class SystemException : public Exception
{
    String_var reason_; // ORBacus specific - not transmitted in
                              // remote method calls
    ULong minor_;
    CompletionStatus status_;

public:

    //
    // Standard IDL to C++ Mapping
    //
    SystemException(const char* = "");
    SystemException(ULong, CompletionStatus);
    SystemException(const char*, ULong, CompletionStatus);
    SystemException(const SystemException&);
    SystemException& operator=(const SystemException&);

    ULong minor() const { return minor_; }
    void minor(ULong minor) { minor_ = minor; }
 
    CompletionStatus completed() const { return status_; }
    void completed(CompletionStatus status) { status_ = status; }

    static SystemException* _downcast(Exception* p)
    { return dynamic_cast<SystemException*>(p); }
    static const SystemException* _downcast(const Exception* p)
    { return dynamic_cast<const SystemException*>(p); }

    //
    // Additional ORBacus specific functions
    //
    const char* reason() const { return reason_; }
    void reason(const char* reason) { reason_ = reason; }

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    // _OB_typeId() only needed for system exceptions, but not for
    // user exceptions
    virtual const char* _OB_typeId() const = 0;

    static void _OB_marshal(const SystemException&, OB::OutputStream&);
    static void _OB_marshalCount(const SystemException&, OB::CountStream&);
    static SystemException* _OB_unmarshal(OB::InputStream&);
    static void _OB_unmarshalThrow(OB::InputStream&);
};

class UserException : public Exception
{
public:

    //
    // Standard IDL to C++ Mapping
    //
    UserException() { }
    UserException(const UserException&) { }
    UserException& operator=(const UserException&)
    { return *this; }

    static UserException* _downcast(Exception* p)
    { return dynamic_cast<UserException*>(p); }
    static const UserException* _downcast(const Exception* p)
    { return dynamic_cast<const UserException*>(p); }
};

#define OB_MAKE_STANDARD_EXCEPTION(T) \
\
class T : public SystemException \
{ \
public: \
\
    T(const char* r = ""); \
    T(ULong, CompletionStatus); \
    T(const char*, ULong, CompletionStatus); \
    T(const T&); \
\
    virtual void _raise() const { throw *this; } \
    static T* _downcast(Exception* p) \
    { return dynamic_cast<T*>(p); } \
    static const T* _downcast(const Exception* p) \
    { return dynamic_cast<const T*>(p); } \
    virtual Exception* _OB_clone() const; \
    virtual void _OB_insertInAny(Any&); \
    virtual const char* _OB_typeId() const; \
};

OB_MAKE_STANDARD_EXCEPTION(UNKNOWN)
OB_MAKE_STANDARD_EXCEPTION(BAD_PARAM)
OB_MAKE_STANDARD_EXCEPTION(NO_MEMORY)
OB_MAKE_STANDARD_EXCEPTION(IMP_LIMIT)
OB_MAKE_STANDARD_EXCEPTION(COMM_FAILURE)
OB_MAKE_STANDARD_EXCEPTION(INV_OBJREF)
OB_MAKE_STANDARD_EXCEPTION(NO_PERMISSION)
OB_MAKE_STANDARD_EXCEPTION(INTERNAL)
OB_MAKE_STANDARD_EXCEPTION(MARSHAL)
OB_MAKE_STANDARD_EXCEPTION(INITIALIZE)
OB_MAKE_STANDARD_EXCEPTION(NO_IMPLEMENT)
OB_MAKE_STANDARD_EXCEPTION(BAD_TYPECODE)
OB_MAKE_STANDARD_EXCEPTION(BAD_OPERATION)
OB_MAKE_STANDARD_EXCEPTION(NO_RESOURCES)
OB_MAKE_STANDARD_EXCEPTION(NO_RESPONSE)
OB_MAKE_STANDARD_EXCEPTION(PERSIST_STORE)
OB_MAKE_STANDARD_EXCEPTION(BAD_INV_ORDER)
OB_MAKE_STANDARD_EXCEPTION(TRANSIENT)
OB_MAKE_STANDARD_EXCEPTION(FREE_MEM)
OB_MAKE_STANDARD_EXCEPTION(INV_IDENT)
OB_MAKE_STANDARD_EXCEPTION(INV_FLAG)
OB_MAKE_STANDARD_EXCEPTION(INTF_REPOS)
OB_MAKE_STANDARD_EXCEPTION(BAD_CONTEXT)
OB_MAKE_STANDARD_EXCEPTION(OBJ_ADAPTER)
OB_MAKE_STANDARD_EXCEPTION(DATA_CONVERSION)
OB_MAKE_STANDARD_EXCEPTION(OBJECT_NOT_EXIST)
OB_MAKE_STANDARD_EXCEPTION(TRANSACTION_REQUIRED)
OB_MAKE_STANDARD_EXCEPTION(TRANSACTION_ROLLEDBACK)
OB_MAKE_STANDARD_EXCEPTION(INVALID_TRANSACTION)
OB_MAKE_STANDARD_EXCEPTION(INV_POLICY)

//
// The Bounds exception
//
struct Bounds : public UserException
{
    //
    // Standard IDL to C++ Mapping
    //
    Bounds() { }
    Bounds(const Bounds& ex) : UserException(ex) { }
    Bounds& operator=(const Bounds&) { return *this; }
    
    static Bounds* _downcast(Exception* p)
    { return dynamic_cast<Bounds*>(p); }
    static const Bounds* _downcast(const Exception* p)
    { return dynamic_cast<const Bounds*>(p); }
    virtual void _raise() const { throw *this; }
    
    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    virtual Exception* _OB_clone() const
    { return new Bounds(*this); }
    virtual void _OB_insertInAny(Any&)
    { assert_nca(false, OB::NCANoExceptionTC); }
};

//
// The UnknownUserException exception
//
/* UnknownUserException is in DII.h */

} // End of namespace CORBA


namespace OB
{

//
// Minor error codes for BAD_PARAM
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java. Furthermore,
// OB::ExceptionToString must be updated.
//
// Note: Minor codes 2-5 are standardized for use by the IR.
//
enum MinorBadParam
{
    MinorDuplicateFormalParam = 6,   // Duplicate formal parameter
    MinorInvalidValueInheritance,    // Invalid inheritance relationship
    MinorInvalidValueModifier,       // Improper valuetype modifier
    MinorDuplicateValueInit,         // Duplicate valuetype initializer
    MinorAbstractValueInit,          // Abstract valuetype cannot have
                                     // initializers
    MinorDuplicateBaseType           // Base type appears more than once
};

//
// Minor error codes for COMM_FAILURE
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java. Furthermore,
// OB::ExceptionToString must be updated.
//
enum MinorCommFailure
{
    MinorRecv = 1,       // recv() failed
    MinorSend,           // send() failed
    MinorRecvZero,       // recv() returned zero
    MinorSendZero,       // send() returned zero
    MinorSocket,         // socket() failed
    MinorSetsockopt,     // setsockopt() failed
    MinorGetsockopt,     // getsockopt() failed
    MinorBind,           // bind() failed
    MinorListen,         // listen() failed
    MinorConnect,        // connect() failed
    MinorAccept,         // accept() failed
    MinorSelect,         // select() failed

    MinorGethostname,    // gethostname() failed
    MinorGethostbyname,  // gethostbyname() failed

    MinorWSAStartup,     // WSAStartup() failed
    MinorWSACleanup,     // WSACleanup() failed

    MinorNoGIOP,         // Not a GIOP message
    MinorUnknownMessage, // Unknown GIOP message
    MinorWrongMessage,   // Wrong GIOP message
    MinorCloseConnection,// Got a close connection message
    MinorMessageError    // Got a message error message
};

//
// Minor error codes for INTF_REPOS
//
// Note: If something is changed here, it must also be changed in
// other ORBacus implementations, like ORBacus for Java. Furthermore,
// OB::ExceptionToString must be updated.
//
enum MinorIntfRepos
{
    MinorNoIntfRepos = 1,       // Interface repository is not available
    MinorLookupAmbiguous,       // Search name for lookup() is ambiguous
    MinorIllegalRecursion       // Illegal recursion
};

//
// Minor error codes for MARSHAL
//
// Note: If something is changed here, it must also be changed in other
// ORBacus implementations, like ORBacus for Java.
//
enum MinorMarshal
{
    MinorReadNoValueFactory = 1,      // Standard minor code for MARSHAL
    MinorReadOverflow,                // Input stream buffer overflow
    MinorReadBooleanOverflow,         // Overflow while reading boolean
    MinorReadCharOverflow,            // Overflow while reading char
    MinorReadWCharOverflow,           // Overflow while reading wchar
    MinorReadOctetOverflow,           // Overflow while reading octet
    MinorReadShortOverflow,           // Overflow while reading short
    MinorReadUShortOverflow,          // Overflow while reading ushort
    MinorReadLongOverflow,            // Overflow while reading long
    MinorReadULongOverflow,           // Overflow while reading ulong
    MinorReadLongLongOverflow,        // Overflow while reading longlong
    MinorReadULongLongOverflow,       // Overflow while reading ulonglong
    MinorReadFloatOverflow,           // Overflow while reading float
    MinorReadDoubleOverflow,          // Overflow while reading double
    MinorReadLongDoubleOverflow,      // Overflow while reading longdouble
    MinorReadStringOverflow,          // Overflow while reading string
    MinorReadStringZeroLength,        // Encountered zero-length string
    MinorReadStringNullChar,          // Encountered null char in string
    MinorReadWStringOverflow,         // Overflow while reading string
    MinorReadWStringZeroLength,       // Encountered zero-length wstring
    MinorReadWStringNullWChar,        // Encountered null wchar in wstring
    MinorReadFixedOverflow,           // Overflow while reading wstring
    MinorReadFixedInvalid,            // Invalid encoding for fixed value
    MinorReadBooleanArrayOverflow,    // Overflow while reading boolean
    MinorReadCharArrayOverflow,       // Overflow while reading char
    MinorReadWCharArrayOverflow,      // Overflow while reading wchar
    MinorReadOctetArrayOverflow,      // Overflow while reading octet
    MinorReadShortArrayOverflow,      // Overflow while reading short
    MinorReadUShortArrayOverflow,     // Overflow while reading ushort
    MinorReadLongArrayOverflow,       // Overflow while reading long
    MinorReadULongArrayOverflow,      // Overflow while reading ulong
    MinorReadLongLongArrayOverflow,   // Overflow while reading longlong
    MinorReadULongLongArrayOverflow,  // Overflow while reading ulonglong
    MinorReadFloatArrayOverflow,      // Overflow while reading float
    MinorReadDoubleArrayOverflow,     // Overflow while reading double
    MinorReadLongDoubleArrayOverflow, // Overflow while reading longdouble
    MinorReadInvTypeCodeIndirection,  // Invalid TypeCode indirection
    MinorWriteOverflow,               // Output stream buffer overflow
    MinorWriteBooleanOverflow,        // Overflow while writing boolean
    MinorWriteCharOverflow,           // Overflow while writing char
    MinorWriteWCharOverflow,          // Overflow while writing wchar
    MinorWriteWCharInvalid,           // Invalid wchar
    MinorWriteOctetOverflow,          // Overflow while writing octet
    MinorWriteShortOverflow,          // Overflow while writing short
    MinorWriteUShortOverflow,         // Overflow while writing ushort
    MinorWriteLongOverflow,           // Overflow while writing long
    MinorWriteULongOverflow,          // Overflow while writing ulong
    MinorWriteLongLongOverflow,       // Overflow while writing longlong
    MinorWriteULongLongOverflow,      // Overflow while writing ulonglong
    MinorWriteFloatOverflow,          // Overflow while writing float
    MinorWriteDoubleOverflow,         // Overflow while writing double
    MinorWriteLongDoubleOverflow,     // Overflow while writing longdouble
    MinorWriteStringOverflow,         // Overflow while writing string
    MinorWriteWStringOverflow,        // Overflow while writing string
    MinorWriteFixedOverflow,          // Overflow while writing wstring
    MinorWriteBooleanArrayOverflow,   // Overflow while writing boolean
    MinorWriteCharArrayOverflow,      // Overflow while writing char
    MinorWriteWCharArrayOverflow,     // Overflow while writing wchar
    MinorWriteOctetArrayOverflow,     // Overflow while writing octet
    MinorWriteShortArrayOverflow,     // Overflow while writing short
    MinorWriteUShortArrayOverflow,    // Overflow while writing ushort
    MinorWriteLongArrayOverflow,      // Overflow while writing long
    MinorWriteULongArrayOverflow,     // Overflow while writing ulong
    MinorWriteLongLongArrayOverflow,  // Overflow while writing longlong
    MinorWriteULongLongArrayOverflow, // Overflow while writing ulonglong
    MinorWriteFloatArrayOverflow,     // Overflow while writing float
    MinorWriteDoubleArrayOverflow,    // Overflow while writing double
    MinorWriteLongDoubleArrayOverflow,// Overflow while writing longdouble
    MinorWriteObjectLocal,            // Attempt to marshal a l-c object
    MinorCountObjectLocal,            // Attempt to count a l-c object
    MinorCountWCharInvalid,           // Invalid wchar
    MinorLongDoubleNotSupported       // long double not supported
};

} // End of namespace OB

#endif
